# Copyright (C) 2004 by Dr. Dieter Maurer, Eichendorffstr. 23, D-66386 St. Ingbert, Germany
# see "LICENSE.txt" for details
#       $Id: IBTrees.py,v 1.3 2005/03/18 15:06:16 dieter Exp $
'''BTrees as incremental searches.'''


from BTrees.OOBTree import OOBTree, OOBucket, OOSet, OOTreeSet
from BTrees.IOBTree import IOBTree, IOBucket, IOSet, IOTreeSet
from BTrees.OIBTree import OIBTree, OIBucket, OISet, OITreeSet
from BTrees.IIBTree import IIBTree, IIBucket, IISet, IITreeSet


from ISearch import HIT, CANDIDATE, END, ISearch, IAnd
from Enumerator import Enumerator

# node positions
CURRENT, KEYS, CHILDREN, NEXT = range(4)

class IBTree(ISearch):
  '''Incremental BTree.

  An 'IBTree' essentially represents a path from
  the root to a key in a leaf.

  The path is implemented as a stack with one entry per level.

  For each level we maintain (in a list)

   * CURRENT, the current position

   * KEYS, the list of keys

   * CHILDREN, the list of children or 'None' (for a leaf)

   * NEXT, the next leaf pointer or 'None' (for an internal node)

  'advanceTo' and 'advanceFrom' usually have logarithic time
  complexity. The case to access the next element is
  optimized and has constant time complexity.
  '''

  # direct reference to the current leaf; 'None' if unknown
  _currentLeaf = None
  _currentLeafId = None # to avoid redundant "nodify"s

  # unnodified current leaf candidate
  _candidateLeaf = None

  # size estimate
  _estimatedSize = None

  # path from root to leaf
  _path = None

  # flatness indication
  _isDeep = None

  # last classification result
  _cr = None

  def __init__(self, btree):
    self._tree = btree

  def getTree(self): return self._tree

  def _setup(self):
    root = self._nodify(self._tree)
    self._isDeep = bool(root[CHILDREN])
    self._path = path = [root]

  def getEstimatedSize(self):
    size = self._estimatedSize
    if size is not None: return size
    path = self._path
    if path is None:
      # setup path
      self._setup(); path = self._path; node = path[0]
      while node[CHILDREN]:
        node = self._nodify(node[CHILDREN][0])
        path.append(node)
    rootSize = len(path[0][KEYS])
    depth = len(path)
    if depth > 1:
      els, ens = _estimatedSizes(self._tree)
      keysPerChild = els * ens ** (depth-2)
      size = (rootSize+1) * keysPerChild
    else: size = rootSize
    self._estimatedSize = size
    return size

  def advanceTo(self, end):
    '''ensure *end* <= value <= min {hit >= *end*}

    return HIT if *end* == value
           CANDIDATE if *end* < value
           END if {hit >= *end*} empty
    '''
    return self._locate(end)


  def advanceFrom(self, start, upto):
    '''ensure *start* < value <= min{hit > *start*} and
       min{ hit > *start*} < *upto* ==> value = min{hit > *start*}

       return 'HIT' if value < *upto*
              'CANDIDATE' if value >= *upto*
              'END' if min{hit > *start*} empty
    '''
    return self._locate(start, False, upto)


  def _getCurrentLeaf(self):
    cl = self._currentLeaf
    if cl is not None: return cl
    cl = self._candidateLeaf
    if cl is None: return cl
    self._candidateLeaf = None
    return self._nodify(cl)

    


  def _locate(self, key, include=True, limit=None):
    '''advance the tree iterator.

    If *include*, we try to locate the least *k* >= *key*;
    otherwise, the least *k* > *key*.

    If we fail, we return 'END'; otherwise we set '_value'
    to the found value and return either 'HIT' or 'CANDIDATE'.
    If *include*, we return 'HIT' iff *key* is in the tree;
    otherwise, if the found *k* < *limit*.
    If we return 'HIT', we advance to the next element.
    '''
    # usage checks
    if __debug__ != 0 and debug:
      value = self._value
      assert key is not None or value is None
      if value is not None:
        cr = cmp(key, value)
        assert cr >= 0
        if include and self._cr == HIT: assert cr > 0
        


    cr = -1

    if self._isDeep:
      # quick check whether the current leaf is already positioned
      # to optimize for the frequent case of lock step trees.
      node = self._getCurrentLeaf()
      if node is not None:
        current, keys, children, next = node; n = len(keys)
        k = keys[current]
        cr = key is None and CANDIDATE or cmp(k, key)
        if cr == HIT and not include:
          # advance to next element
          current += 1; node[CURRENT] = current
          if current >= n:
            # advance to next leaf
            if next is None: return END
            node = self._nodify(next)
            current, keys, children, next = node; n = len(keys)

    if cr < 0:
      # current leaf not already positioned -- traverse from root
      i = 0; path = self._path
      while True:
        if path is None: self._setup(); path = self._path
        if i == len(path):
          current, _, children = path[-1][:3]
          node = self._nodify(children[current])
          path.append(node)
        else: node = path[i]
        current, keys, children, next = node
        n = len(keys)
        # "target" is 1, when we need "k > key", and 0 for "k >= key"
        target = int(bool(children) or not include)
        if current >= n:
          if not children: break
          i += 1
          continue
        # perform binary search
        # let k := keys[current]; keys[n] := infinity
        cr = cmp(keys[current], key)
        if cr >= target:
          if not children: d = cr; break
          i += 1
          continue
        # cmp(k, key) < target
        # widening loop
        d = 1
        while True:
          # cmp(k, key) < target
          d = min(d, n-1-current)
          if d == 0:
            # current = n-1
            current = n;
            # cmp(k, key) >= target; cmp(keys[current-1], key) < target
            # current == n
            break
          cr = cmp(keys[current + d], key)
          if cr == HIT:
            current += d; d = 0
            if target: current += 1
            # d == 0 and cmp(k, key) >= target; cmp(keys[current-1], key) < target
            # not target => cmp(k, key) == 0
            break
          if cr == CANDIDATE: break
          current += d; d <<= 1
          # cmp(k, key) < target

        # d == 0 => cmp(k, key) >= target; cmp(keys[current-1], key) < target
        # not target => (current == n or cmp(k, key) == 0) iff d == 0
        # d != 0 => cmp(k, key) < target; cmp(keys[current+d], key) >= target

        # narrowing loop
        while d > 1:
          # d >= 2 and cmp(k, key) < target; cmp(keys[current+d], key) >= target
          mid = d // 2
          cr = cmp(keys[current + mid], key)
          if cr == HIT:
            current += mid; d = 0
            if target: current += 1
            # d == 0 and cmp(k, key) >= target; cmp(keys[current-1], key) < target
            # not target => cmp(k, key) == 0
            break
          elif cr > 0: d = mid
          else: current += mid; d -= mid
          # cmp(k, key) < target; cmp(keys[current+d], key) >= target

        # 0 <= d <= 1
        # d == 0: cmp(k, key) >= target; cmp(keys[current-1], key) < target
        # d == 1: cmp(k, key) < target; cmp(keys[current+d], key) >= target
        # not target => (current == n or cmp(k, key) == 0) iff d == 0
        current += d
        # cmp(k, key) >= target; cmp(keys[current-1], key) < target
        # not target => (current == n or cmp(k, key) == 0) iff d == 0

        if not children: break

        # update node
        node[CURRENT] = current
        children[current] = node = self._nodify(children[current])
        i += 1
        path[i:] = [node]

      # leaf level
      # cmp(k, key) >= target; cmp(keys[current-1], key) < target
      # target == not include
      # include => (current == n or cmp(k, key) == 0) iff d == 0
      if current >= n:
        # not in this bucket
        # the interesting element may be in the next bucket
        if next is None: return END # no more hits
        (current, keys, children, next) = self._nodify(next)
        d = 1 # indicate CANDIDATE


      # include => cmp(k, key) == 0 iff d == 0
      if include and d == 0: cr = HIT
      else: cr = CANDIDATE

    # save value
    self._value = k = keys[current]
    # classify 
    if not include:
      if limit is None or cmp(k, limit) < 0: cr = HIT
      else: cr = CANDIDATE
    # pass over hit
    if cr == HIT:
      current += 1
    node[CURRENT] = current
    if current >= n:
      self._currentLeaf = self._currentLeafId = None
      self._candidateLeaf = next
    self._cr = cr
    return cr


  def _nodify(self, tree):
    '''convert *tree* into a node
     
    A node is a 4 element list; see "node positions" above.
    '''
    tid = id(tree)
    if tid == self._currentLeafId: return self._currentLeaf # optimization
    state = tree.__getstate__()
    if not state: state = (),
    content = state[0]
    ref = len(state) > 1 and state[1] or None
    withValues, isLeaf = _classify(tree)
    if ref is None and not isLeaf:
      # leaf inside node optimization
      isLeaf = True
      # "not content" means an empty tree -- not sure why an empty tree is not classified as "leaf"
      if content: content = content[0][0]
    n = len(content)
    if isLeaf:
      children = None
      if withValues: keys = [content[i] for i in range(0, n, 2)]
      else: keys = content
    else:
      children = [content[i] for i in range(0, n, 2)]
      keys = [content[i] for i in range(1, n, 2)]
    node = [0, keys, children, ref]
    if isLeaf:
      self._currentLeafId = tid
      self._currentLeaf = node
    return node


# BTrees as Enumerator
class EBTree(Enumerator):
  '''BTree as an enumerator'''
  def __init__(self, btree):
    self._tree = btree

  def check(self, elem):
    return self._tree.has_key(elem)

  def next(self, elem):
    candidates = self._tree.keys(elem)
    # from ZODB 3.3 on, we can specify that we do not want "elem"
    if not candidates: return
    candidate = candidates[0]
    if cmp(candidate, elem) > 0: return candidate
    try: return candidates[1]
    except IndexError: return


# classification: type -> (maxLeafSize, maxNodeSize, withValues)
# size values from C code
_classifier = {
  type(OOBTree()) : (30, 250, True,),
  type(IOBTree()) : (60, 500, True,),
  type(OIBTree()) : (60, 250, True,),
  type(IIBTree()) : (120, 500, True,),
  type(OOBucket()) : (None, None, True,),
  type(IOBucket()) : (None, None, True,),
  type(OIBucket()) : (None, None, True,),
  type(IIBucket()) : (None, None, True,),
  type(OOTreeSet()) : (30, 250, False,),
  type(IITreeSet()) : (30, 250, False,),
  type(OOSet()) : (None, None, False,),
  type(IISet()) : (None, None, False,),
  }

def _classify(tree):
  '''-> *withValue*, *isLeaf.'''
  mls, mns, withValues = _classifier[type(tree)]
  return withValues, mls is None

def _estimatedSizes(tree):
  '''the pair of estimated sizes for leafs and nodes.'''
  mls, mns, withValues = _classifier[type(tree)]
  if mls is None: return mls, mns
  else:
    # we assume that inserted items are often sorted and
    # that therefore nodes/leafs are smaller than expected
    return 0.65 * mls, 0.6 * mns

def _getSetType(tree):
  '''the set type corresponting to *tree*.'''
  return type(tree).__name__.startswith('I') and IISet or OOSet

  
def intersection(*trees):
  '''intersect all *trees*.'''
  # filter 'None' to be compatible with the standard intersection
  # Note, that we convert type even for a unary intersection
  trees = [tree for tree in trees if tree is not None]
  if not trees: raise ValueError('empty intersection')
  target = _getSetType(trees[0])()
  search = IAnd(* map(IBTree, trees)); search.complete()
  target.__setstate__((tuple(search),))
  return target



# debugging control
debug = 1
