# Copyright (C) 2004-2006 by Dr. Dieter Maurer, Eichendorffstr. 23, D-66386 St. Ingbert, Germany
# see "LICENSE.txt" for details
#       $Id: ISearch.py,v 1.8 2006/04/09 16:52:49 dieter Exp $
'''IncrementalSearch is a framework to perform incremental 'and', 'or',
'not' and other query types.
'''

from heapq_select import heappop, heappush, heapreplace

##Codes
HIT = 0 # current value is a hit
NOHIT = 2 # current value is not a hit
CANDIDATE = 1 # current value is the next candidate for a hit (but may in fact
              # not be a hit
END = 3 # there are no more hits

class ISearch:
  '''abstract base class representing an increasing sequence
  and a current value (of the same element type).

  Elements in the sequence are called hits.

  The current value strictly increases.
  '''
  _value = _cr = None

  def getValue(self):
    '''the current value'''
    return self._value

  def getClassification(self):
    '''the classification of the current value.'''
    return self._cr

  def getEstimatedSize(self):
    raise NotImplementedError

  def advanceTo(self, end):
    '''ensure *end* <= value <= min {hit >= *end*}.
    Return 'HIT', when *end* is a hit.
    '''
    raise NotImplementedError

  def advanceFrom(self, start, upto):
    '''ensure *start* < value <= min{hit > *start*}
    and either value is a hit or value >= *upto*.

    Return 'HIT' when value is a hit and < *upto*.

    *start* and/or *upto* may be 'None' meaning no restriction.
    '''
    raise NotImplementedError

  # iteration protocol: note that we must iterate at most onces!
  def __iter__(self):
    '''a one time iterator -- note that we are not safe against misuse.'''
    if self._value is not None:
      raise ValueError('can iterate at most once')
    return self

  def next(self):
    cr = self.advanceFrom(self._value, None)
    if cr == END: raise StopIteration
    return self._value


class MultiSearch(ISearch):
  '''a combined search over one or more ISearches.'''
  _completed = False

  def __init__(self, *searches):
    self._searches = list(searches)

  def add(self, *searches):
    '''add more searches'''
    if self._completed:
      raise ValueError('must not add elements to a completed MultiSearch')
    self._searches += list(searches)

  def complete(self):
    '''used for optimizations -- not yet implemented'''
    self._completed = True


class IAnd(MultiSearch):
  '''the 'and' (intersection) of the associated searches.'''
  def advanceTo(self, end):
    assert self._completed
    for s in self._searches:
      cl = s.advanceTo(end)
      if cl != HIT: break
    if cl in (HIT, CANDIDATE): self._value = s.getValue()
    else: self._value = end
    self._cr = cl
    return cl

  def advanceFrom(self, start, upto):
    assert self._completed
    searches = self._searches
    while True: # search loop
      s = searches[0]
      cl = s.advanceFrom(start, upto)
      if cl in (END, NOHIT): return cl
      value = self._value = s.getValue()
      self._cr = cl
      if cl != HIT: return cl
      provider = s
      while True: # verification loop
        for s in searches: # check round
          if s is provider: provider = None; continue
          cl = s.advanceTo(value)
          if cl != HIT: break
        if cl == NOHIT: start = value; break # restart with "value"
        if cl == END: return cl
        value = self._value = s.getValue()
        self._cr = cl
        if cl == CANDIDATE:
            if not (upto is None or value < upto): return cl
            # continue verification with new value
        else: return cl # HIT

  def complete(self):
    if self._completed: return
    MultiSearch.complete(self)
    searches = self._searches
    if not searches:
      raise ValueError('cannot handle empty and searches')
    # merge in nested "IAnd"s
    ns = []
    for s in searches:
      if isinstance(s, IAnd):
        assert s._completed
        ns += s._searches
      else: ns.append(s)
    # order by increasing size -- for efficiency
    # Note, that later we may want more criteria for this sorting
    proper = []; improper = []
    for s in ns:
      size = s.getEstimatedSize()
      if size is None: improper.append(s)
      else: proper.append((size, s))
    proper.sort()
    self._searches = [s[1] for s in proper] + improper

  def getEstimatedSize(self):
    '''we use the minimal value of the subqueries estimates -- this is above
    the true value.
    '''
    sizes = [s.getEstimatedSize() for s in self._searches]
    sizes = [size for size in sizes if size is not None]
    if not sizes: return None
    return min(sizes)


class IOr(MultiSearch):
  '''the 'or' (union) of the associated searches.'''
  def advanceTo(self, end):
    assert self._completed
    heap = self._heap
    self._setNonCandidateBoundary()
    cl = -1
    # update heap until either we find a hit or all values are > *end*.
    while heap:
      s = heap[0]
      cr = cmp(s.getValue(), end)
      if cr > 0: break
      if cr == 0 and s.getClassification() == HIT: cl = HIT; break
      clt = s.advanceTo(end)
      if clt <= CANDIDATE:
        # Note: we could push a hit to "non candidates"
        #   as the value is useless. However, "heapreplace" has
        #   the same costs as "heappop". Thus, we trade
        #   2 "heapreplace" against a "heappop" and a "heappush"
        heapreplace(heap, s)
        if clt == HIT: cl = HIT; break
      else:
        heappop(heap)
        if clt == NOHIT: self._pushNonCandidate(s)
    if cl == -1:
      # check non candidates
      while True:
        s = self._popNonCandidate()
        if s is None: break
        cl = s.advanceTo(end)
        if cl <= CANDIDATE:
          heappush(heap, s)
          if cl == HIT: break
        elif cl == NOHIT: self._pushNonCandidate(s)
      if cl != HIT:
        # we did not find a hit
        if self._pushedNonCandidate():
          # we have non candidates
          self._cr = NOHIT
          return NOHIT
        if heap: cl = CANDIDATE; s = heap[0]
        else: return END
    self._cr = cl; self._value = s.getValue()
    return cl

  def advanceFrom(self, start, upto):
    assert self._completed
    self._setNonCandidateBoundary()
    heap = self._heap
    # update heap until it only contains values > "start" and
    # the bottom element is a potential candidate
    limit = upto
    cl = -1
    while heap:
      s = heap[0]
      value = s.getValue()
      # Note: when something is in the heap, *start* will not be None
      cr = cmp(value, start)
      if cr > 0:
        if cl == -1 and upto is not None and cmp(value, upto) >= 0:
          cl = CANDIDATE; break
        clt = s.getClassification()
        unchanged = True
        if clt != HIT:
          clt = s.advanceTo(value)
          # Note: this check was missing
          if clt == NOHIT:
            # skip this intervening value
            # Note: we cannot use "advanceFrom" when we already reached
            #  "limit"; in this case, however, "cl = HIT" and, "limit"
            #  is a hit and "value == limit"
            if limit is not None and cmp(value, limit) >= 0: clt = HIT
            else:
              unchanged = False
              clt = s.advanceFrom(value, limit)
              value = s.getValue()
        if unchanged and clt == HIT: limit = value; cl = HIT; break
      else: clt = s.advanceFrom(start, limit)
      if clt <= CANDIDATE:
        heapreplace(heap, s)
        # Note: any hit is better than a previous one
        if clt == HIT: limit = s.getValue(); cl = HIT;
      else:
        heappop(heap)
        if clt == NOHIT: self._pushNonCandidate(s)
    # process non candidates
    while True:
      s = self._popNonCandidate()
      if s is None: break
      clt = s.advanceFrom(start, limit)
      if clt <= CANDIDATE:
        heappush(heap, s)
        # Note: any hit is better than a previous one
        if clt == HIT: limit = s.getValue(); cl = HIT
        elif cl == -1: cl = CANDIDATE
      elif clt == NOHIT: self._pushNonCandidate(s)
    if cl == -1:
      # we did not find a hit nor a candidate -- especially, the heap is empty
      if self._pushedNonCandidate(): self._cr = NOHIT; return NOHIT
      return END
    if cl != HIT:
      # cl == CANDIDATE
      if self._pushedNonCandidate(): self._cr = NOHIT; return NOHIT # cannot trust
      value = heap[0].getValue()
    else: value = limit
    self._value = value; self._cr = cl
    return cl
      
  def complete(self):
    if self._completed: return
    MultiSearch.complete(self)
    self._heap = []
    ns = []
    # merge in nested IOrs
    for s in self._searches:
      if isinstance(s, IOr):
        assert s._completed
        ns += [s.unwrap() for s in s._searches if s is not None]
      else: ns.append(s)
    # order by decreasing size -- for efficiency
    # Note, that later we may want more criteria for this sorting
    proper = []; improper = []
    for s in ns:
      size = s.getEstimatedSize()
      if size is None: improper.append(s)
      else: proper.append((size, s))
    proper.sort(); proper.reverse()
    self._setupNonCandidates(
      [_OrderWrapper(s[1]) for s in proper] + map(_OrderWrapper, improper)
      )

  ## "non_candidates" handling -- i.e. searches, we do not know a candidate
  ## for. Currently, it is maintained as a ring, abusing "_searches"
  ## Order is not maintained.
  ## Note: "_non_candidate" handling can be optimized
  ##   See "HitListClassifiedList"
  def _setupNonCandidates(self, searches):
    self._non_candidates = self._searches = searches
    self._popindex = 0; self._pushindex = len(searches)
    # to avoid boundary problems
    searches.append(None)
    self._ringsize = len(searches)

  def _setNonCandidateBoundary(self):
    self._boundary = self._pushindex

  def _popNonCandidate(self):
    '''the next non candidate or 'None'.'''
    pi = self._popindex; rs = self._ringsize; ncs = self._non_candidates
    if pi == self._boundary: return
    nc = ncs[pi]; ncs[pi] = None
    pi += 1
    if pi == rs: pi = 0
    self._popindex = pi
    return nc

  def _pushNonCandidate(self, nc):
    pi = self._pushindex; rs = self._ringsize; ncs = self._non_candidates
    ncs[pi] = nc
    pi += 1
    if pi == rs: pi = 0
    self._pushindex = pi

  def _pushedNonCandidate(self):
    '''true, when we pushed a least one non candidate.'''
    return self._boundary != self._pushindex

    

  def getEstimatedSize(self):
    '''we use the sum of the subqueries estimates -- this is above
    the true value.

    We do not count 'None' sizes hoping they will not dominate.
    '''
    sizes = [s.getEstimatedSize() for s in self._searches if s is not None]
    if not sizes: return 0
    sizes = [size for size in sizes if size is not None]
    if not sizes: return None
    return sum(sizes)

  def _remove(self, remove):
    '''remove searches identified by sorted *remove*.'''
    searches = self._searches
    i = 0
    for s in remove:
      while True:
        if searches[i] is s: del searches[i]; break
        i += 1



class _OrderWrapper(ISearch):
  __slots__ = ('__search', '_cr', '_value')

  def __init__(self, search):
    self.__search = search

  def unwrap(self): return self.__search

  def advanceTo(self, end):
    search = self.__search
    cr = self._cr = search.advanceTo(end)
    self._value = search.getValue()
    return cr

  def advanceFrom(self, start, upto):
    search = self.__search
    cr = self._cr = search.advanceFrom(start, upto)
    self._value = search.getValue()
    return cr

  def getEstimatedSize(self):
    return self.__search.getEstimatedSize()

  def __cmp__(self, other):
    return cmp(self._value, other._value) or cmp(self._cr, other._cr)


class INot(ISearch):
  '''incremental Not.'''
  _search = None

  def __init__(self, search, enumerator=None):
    self._search = search
    self._enumerator = enumerator

  def advanceTo(self, end):
    '''we assume without check that *end* is in the domain.'''
    search = self._search
    self._value = end
    if search is None: self._cr = HIT; return HIT
    # Note: this can be simplified when we drop the "assert" in "IBTree".
    candidate = self._candidate; cl = None
    if candidate is not None:
      if cmp(end, candidate) < 0: cl = NOHIT
      else: del self._candidate
    if cl is None: cl = search.advanceTo(end)
    if cl == END: del self._search
    elif cl == CANDIDATE: self._candidate = search.getValue()
    cl = self._cr = cl == HIT and NOHIT or HIT
    return cl

  _candidate = None
  def advanceFrom(self, start, upto):
    search = self._search; enumerator = self._enumerator
    next = start
    cl = HIT
    while True:
      next = enumerator.next(next)
      if next is None: self._cr = END; return END
      if upto is not None and cmp(next, upto) >= 0: break
      if search is None: cl = NOHIT; break
      # Note: this can be simplified when we drop the "assert" in "IBTree".
      candidate = self._candidate
      if candidate is not None:
        if cmp(next, candidate) < 0: cl = NOHIT; break
        del self._candidate
      cl = search.advanceTo(next)
      if cl != HIT: break
    if cl == END: del self._search
    elif cl == CANDIDATE: self._candidate = search.getValue()
    self._value = next
    cl = self._cr = cl == HIT and CANDIDATE or HIT
    return cl

  def getEstimatedSize(self):
    '''we do not try to estimate the size but handle 'None' sizes specially.'''
    return None


class IFilter(ISearch):
  '''incremental filter.'''
  _filter = None

  def __init__(self, filter, enumerator=None):
    self._filter = filter
    self._enumerator = enumerator

  def advanceTo(self, end):
    '''we assume without check that *end* is in the domain.'''
    filter = self._filter
    self._value = end
    fr = filter(end)
    if fr: cr = HIT
    else: cr = NOHIT
    self._cr = cr
    return cr

  def advanceFrom(self, start, upto):
    filter = self._filter; enumerator = self._enumerator
    next = start
    cr = NOHIT
    while True:
      next = enumerator.next(next)
      if next is None: self._cr = END; return END
      if upto is not None and cmp(next, upto) >= 0: cr = CANDIDATE; break
      fr = filter(next)
      if fr: cr = HIT; break
    self._value = next
    self._cr = cr
    return cr

  def getEstimatedSize(self):
    '''we do not try to estimate the size but handle 'None' sizes specially.'''
    return None
