# Copyright (C) 2004 by Dr. Dieter Maurer, Eichendorffstr. 23, D-66386 St. Ingbert, Germany
# see "LICENSE.txt" for details
#       $Id: testIBTrees.py,v 1.4 2006/01/18 21:25:16 dieter Exp $
'''Tests for BTrees based incremental searches.'''

from unittest import TestCase, TestSuite, makeSuite, TextTestRunner

import BTrees
from BTrees.OOBTree import OOBTree, OOSet, OOBucket
from BTrees.IIBTree import IISet, IITreeSet

from IncrementalSearch import IBTrees
from IncrementalSearch.IBTrees import IBTree, intersection
from IncrementalSearch.ISearch import HIT, CANDIDATE, END


# activate debugging
IBTrees.debug = True

class BTreeISearchTestBase(TestCase):
  '''test suite controlled by *keys* and *split*.'''

  # to be defined by deriving classes
  keys = None # keys the constructed BTree should have
  split = None # where the tree should have subtrees

  def setUp(self):
    keys = self.keys
    set =  self._set = OOSet(keys)
    tree, _ = self._makeTree(set, self.split)
    self._search = IBTree(tree)

  def _makeTree(self, set, split, next=None):
    keys = list(set.keys())
    n = len(set)
    if split:
      i = 0; children = []
      for s in split:
        grandChildren = []
        while True:
          if i >= n: break
          x = set[i]
          if x >= s: break
          grandChildren.append(x); i += 1
        children.append(grandChildren)
        children.append(s)
      children.append(set.keys(split[-1]))
      for i in range(len(children)-1, -1, -2):
        tree = OOBucket(); items = []
        for k in children[i]: items += [k, None]
        S = (tuple(items),)
        if next is not None: S += (next,)
        tree.__setstate__(S)
        next = children[i] = tree
      tree = OOBTree()
      tree.__setstate__((tuple(children), children[0]))
      return tree, children[0]
    else:
      tree = OOBucket(); items = []
      for k in keys: items += [k, None]
      tree.__setstate__((tuple(items),))
      return tree, None

  def testTraverseFrom(self):
    val = None; search = self._search 
    for k in self.keys:
      cr = search.advanceFrom(val, None)
      self.assertEqual(cr, HIT)
      self.assertEqual(search.getValue(), k)
      val = k
    self.assertEqual(search.advanceFrom(val, None), END)

  def testTraverseTo(self):
    search = self._search 
    for k in self.keys:
      self.assertEqual(search.advanceTo(k), HIT)
      self.assertEqual(search.getValue(), k)

  def testToWithSkipping(self):
    self._check(3, 4, 25, 30, 31)

  def testToBoundaries(self):
    self._check(14, 15, 28, 30, 31, 40, 41)

  def testFromSkipping(self):
    self._check((3,4), 4, (4,25), 5, (25,26), (29,None), (30,None))

  def testFromBoundaries(self):
    self._check((14,15), (15,16), (29,31), (29,36))

  def testFromBoundaries2(self):
    self._check((15,20))

  def _check(self, *cmds):
    current = None; cr = None; set = self._set; search = self._search
    for cmd in cmds:
      args = cmd
      if isinstance(cmd, tuple): val = cmd[0]; fun = search.advanceFrom
      else: val = cmd; fun = search.advanceTo; args = cmd,
      if val is None and current is not None \
         or val < current \
         or val == current and cr == HIT and not isinstance(cmd, tuple):
        self.assertRaises(AssertionError, fun, *args)
      else:
        cr = fun(*args)
        hits = set.keys(val)
        if not hits: self.assertEqual(cr, END); continue
        if isinstance(cmd, tuple):
          if hits[0] == val: hits = hits[1:]
          if not hits: self.assertEqual(cr, END); continue
          else:
            current = hit = hits[0]; upto = cmd[1]
            self.assertEqual(search.getValue(), hit)
            if upto is None or hit < upto: self.assertEqual(cr, HIT)
            else: self.assertEqual(cr, CANDIDATE)
        else:
          current = hit = hits[0]
          self.assertEqual(search.getValue(), hit)
          self.assertEqual(cr, hit > val and CANDIDATE or HIT)

class TestOneLevel(BTreeISearchTestBase):
  keys = range(1,31)

class TestSecondTrivialLevel(TestOneLevel):
  split = range(2,31)

class TestRegularSplit(TestOneLevel):
  slit = (4, 25,)

class TestIrregular(BTreeISearchTestBase):
  keys = range(2,15,2) + range(15,30,2) + [31, 35, 40]
  split = (15, 30,)


class TestTrueTree(BTreeISearchTestBase):
  # to be overridden
  keys = None
  class_ = OOBTree
  mapping = True

  def setUp(self):
    keys = self.keys
    self._set = OOSet(keys)
    if self.mapping: items = map(None, keys, ())
    else: items = keys
    self._search = IBTree(self.class_(items))


class TestMedBTree(TestTrueTree):
  keys = range(1, 70)


class TestMedBTree2(TestTrueTree):
  keys = range(1, 150, 2)

class TestSmallBTree(TestTrueTree):
  keys = range(1,10)

class TestBucket(TestTrueTree):
  keys = range(1, 70)
  class_ = OOBucket

class TestTreeSet(TestTrueTree):
  keys = range(1, 500)
  class_ = IITreeSet
  mapping = False

class TestSet(TestTreeSet):
  class_ = IISet

class TestEmptyTree(TestTrueTree):
  keys = ()

  def testEstimatedSize(self):
    self.assertEqual(self._search.getEstimatedSize(), 0)


class TestDepthThree(BTreeISearchTestBase):
  keys1 = range(2,15,2); split1 = [10]
  keys2 = range(15,30,2); split2 = [19]
  keys3 = [31, 35, 40]; split3 = [38]
  keys = keys1 + keys2 + keys3
  split = [15, 30]

  def setUp(self):
    self._set = OOSet(self.keys1 + self.keys2 + self.keys3)
    mkt = self._makeTree
    tree3, leaf3 = mkt(OOSet(self.keys3), self.split3)
    tree2, leaf2 = mkt(OOSet(self.keys2), self.split2, leaf3)
    tree1, leaf1 = mkt(OOSet(self.keys1), self.split1, leaf2)
    tree = OOBTree(); split = self.split
    tree.__setstate__(((tree1, split[0], tree2, split[1], tree3), leaf1))
    self._search = IBTree(tree)



### Intersection tests
class IntersectionTestBase(TestCase):
  # to be overridden in deriving classes
  class_ = None
  mapping = None
  module = None

  set = range(1000)

  def testEmptyIntersection(self):
    self.assertRaises(ValueError, intersection)

  def testUnaryIntersection(self):
    self._check(1)

  def testEqualIntersection(self):
    self._check(1, 1)

  def testSubsetIntersection(self):
    self._check(2, 1)
    self._check(1, 2)

  def testEmptyResult(self):
    self._check(2, (2,1))

  def testTripleIntersection(self):
    self._check(2, 3, 5)

  def testSmallResult(self):
    s = [100, 200]
    self._check(s, 1)
    self._check(1, s)

  def _check(self, *args):
    args = map(self._makeSet, args)
    result = None; bti = self.module.intersection
    for set in args: result = bti(result, set)
    self.assert_(list(result.keys()) == intersection(*args).keys())

  def _makeSet(self, spec):
    '''make a set from *spec*.

    *spec* can be an integer (meaning: 'spec * self._set')
    a pair (*f*, *o*) (meaning 'f * self._set + o')
    or a list (meaning itself).
    '''
    set = self.set
    if isinstance(spec, int): spec = [spec * x for x in set]
    elif isinstance(spec, tuple):
      f, o = spec
      spec = [f * x + o for x in set]
    if self.mapping: spec = map(None, spec, ())
    return self.class_(spec)


class TestSetIntersection(IntersectionTestBase):
  class_ = IISet
  mapping = False
  module = BTrees.IIBTree

class TestTreeSetIntersection(TestSetIntersection):
  class_ = IITreeSet

class TestBucketIntersection(IntersectionTestBase):
  class_ = OOBucket
  module = BTrees.OOBTree
  mapping = True

class TestBTreeIntersection(TestBucketIntersection):
  class_ = OOBTree
    

def test_suite():
  return TestSuite([makeSuite(cl) for cl in [
                     TestOneLevel,
                     TestSecondTrivialLevel,
                     TestRegularSplit,
                     TestIrregular,
                     TestMedBTree,
                     TestMedBTree2,
                     TestSmallBTree,
                     TestEmptyTree,
                     TestBucket,
                     TestTreeSet,
                     TestSet,
                     TestDepthThree,
                     TestSetIntersection,
                     TestTreeSetIntersection,
                     TestBucketIntersection,
                     TestBTreeIntersection,
                     ]
                    ]
                   )

def main():
  tester = TextTestRunner()
  tester.run(test_suite())

if __name__ == '__main__': main()
