# Copyright (C) 2004-2006 by Dr. Dieter Maurer, Eichendorffstr. 23, D-66386 St. Ingbert, Germany
# see "LICENSE.txt" for details
#       $Id: testIBTrees.py,v 1.1.1.1 2004/08/14 14:27:55 dieter Exp $
'''IncrementalSearch tests.'''

from unittest import TestCase, TestSuite, makeSuite, TextTestRunner
from BTrees.IIBTree import IISet

from IncrementalSearch.IBTrees import IBTree, EBTree
from IncrementalSearch.ISearch import HIT, CANDIDATE, NOHIT, END,\
     IOr, INot, IAnd, ISearch, IFilter

class IOrTests(TestCase):
  def setUp(self):
    self._search = IOr(
      IBTree(IISet(range(5,20,5))),
      IBTree(IISet(range(2,20,2))),
      )
    self._search.complete()

  def testResult(self):
    self.assertEqual(list(self._search),
                     IISet(range(5,20,5) + range(2,20,2)).keys()
                     )
   
  def testToTraversal(self):
    search = self._search
    l = []; i = 0
    while True:
      cr = search.advanceTo(i)
      self.assertEqual(i >= 19, cr == END)
      if cr == END: break
      if cr == HIT: l.append(search.getValue())
      if cr == CANDIDATE: i = search.getValue()
      else: i += 1
    self.assertEqual(l, IISet(range(5,20,5) + range(2,20,2)).keys())

  def testFromJumping(self):
    search = self._search
    for s,r in ((4,5), (6,8), (14,15), (15,16)):
      self.assertEqual(search.advanceFrom(s, None), HIT)
      self.assertEqual(search.getValue(), r)
    self.assertEqual(search.advanceFrom(18, None), END)

  def testFromLimit(self):
    search = self._search
    self.assertEqual(search.advanceFrom(2,4), CANDIDATE)
    self.assertEqual(search.getValue(), 4)
    self.assertEqual(search.advanceFrom(8,11), HIT)
    self.assertEqual(search.getValue(), 10)
    self.assertEqual(search.advanceFrom(14,16), HIT)
    self.assertEqual(search.getValue(), 15)

  def testFromToInterleaving(self):
    search = self._search
    self.assertEqual(search.advanceFrom(8,10), CANDIDATE)
    self.assertEqual(search.getValue(), 10)
    self.assertEqual(search.advanceTo(10), HIT)
    self.assertEqual(search.advanceFrom(10,11), CANDIDATE)
    self.assertEqual(search.getValue(), 12)
    self.assertEqual(search.advanceTo(12), HIT)
    self.assertEqual(search.advanceFrom(12,13), CANDIDATE)
    self.assertEqual(search.getValue(), 14)
    self.assertEqual(search.advanceTo(15), HIT)

class INotTests(TestCase):
  def setUp(self):
    self._search = INot(
      IBTree(IISet(range(4, 12, 4))),
      EBTree(IISet(range(2, 12, 2))),
      )

  def testResult(self):
    self.assertEqual(list(self._search), [2, 6, 10])

  def testToTraversal(self):
    search = self._search
    for i in range(2, 12, 2):
      cr = search.advanceTo(i)
      self.assertEqual(i > 10, cr == END)
      self.assertEqual(i in [2, 6, 10], cr == HIT)
      self.assertEqual(i <= 10 and i not in [2, 6, 10], cr == NOHIT)

  def testFromLimit(self):
    search = self._search
    self.assertEqual(search.advanceFrom(2,6), CANDIDATE)
    self.assertEqual(search.getValue(), 6)
    self.assertEqual(search.advanceTo(6), HIT)
    self.assertEqual(search.getValue(), 6)
    self.assertEqual(search.advanceFrom(7,11), HIT)
    self.assertEqual(search.getValue(), 10)
    self.assertEqual(search.advanceFrom(11,13), END)

class IFilterTests(INotTests):
  def setUp(self):
    self._search = IFilter(
      lambda i, inot=range(4,12,4): i not in inot,
      EBTree(IISet(range(2, 12, 2))),
      )


class IntegrationTests(TestCase):
  def testAndOrNot(self):
    enumerator = EBTree(IISet(range(0,17)))
    search = IOr(
      IBTree(IISet(range(5,17,5))),
      INot(IBTree(IISet(range(5,11))), enumerator),
      )
    search.complete()
    search = IAnd(
      search,
      INot(IBTree(IISet(range(1,17,2))), enumerator),
      )
    search.complete()
    self.assertEqual(list(search), [0,2,4,10,12,14,16])


class _HitNohitISearch(ISearch):
  '''auxiliary class for covering IOr test.

  Does not report candidates but only hits and nohits.
  '''
  _value = _cr = None

  def __init__(self, set):
    self._set = set

  def advanceTo(self, end):
    return self._locate(end)

  def advanceFrom(self, start, upto):
    return self._locate(start, False, upto)

  def _locate(self, key, include=True, limit=None):
    # checking code copied from IBTree
    value = self._value
    assert key is not None or value is None
    if value is not None:
      cr = cmp(key, value)
      assert cr >= 0
      if include and self._cr == HIT: assert cr > 0
    keys = list(self._set.keys(key))
    if not include and key in keys: keys.remove(key)
    if not keys: return END
    if include: cr = key != keys[0] and NOHIT or HIT
    else: cr = limit is not None and keys[0] >= limit and NOHIT or HIT
    self._cr = cr
    if cr == HIT: self._value = keys[0]
    else: self._value = key
    return cr

  def getEstimatedSize(self):
    return len(self._set)

class IOrCoverTests(TestCase):
  def setUp(self):
    s1 = IISet((0, 5, 10))
    s2 = IISet((0,3,6,10,12))
    self._search = IOr(IBTree(s1), _HitNohitISearch(s2))
    self._search.complete()
    s1.update(s2)
    self._result = s1

  def testAdvanceTo(self):
    r = self._result; s = self._search; to = s.advanceTo
    for i in range(r.minKey(), r.maxKey()+1):
      cr = to(i)
      self.assertEqual(cr, i not in r and NOHIT or HIT)
      if cr == HIT: self.assertEqual(s.getValue(), i)
    self.assertEqual(to(r.maxKey()+1), END)

  def testEnumerate(self):
    self.assertEqual(list(self._result), list(self._search))

  def testAdvanceFrom(self):
    s = self._search; fr = s.advanceFrom
    def check(start, upto, result=None):
      cr = fr(start, upto)
      self.assert_((result is not None) == (cr == HIT))
      if cr == HIT: self.assertEqual(result, s.getValue())
    check(1,3)
    check(1,4,3)
    check(3,7,5)
    check(5,7,6)
    check(6,10)
    check(6,13, 10)
    check(11,13, 12)
    self.assertEqual(fr(12,14), END)


class _ISearchCandidate1Hit2(ISearch):

  def getEstimatedSize(self): return -1

  def advanceTo(self, end):
    self.value = end
    cl = end != 2 and NOHIT or HIT
    self.classification = cl
    #print 'advanceTo', end, cl
    return cl

  def advanceFrom(self, fr, upto):
    if fr is None or fr < 1:
      if upto is None or upto > 2: v = 2; cl = HIT
      elif upto <= 1: v = 1; cl = CANDIDATE
      else: v = 2; cl = CANDIDATE
    elif fr < 2:
      if upto is None or upto > 2: v = 2; cl = HIT
      else: v = upto; cl = CANDIDATE
    else: cl = END; v = 2
    self.classification = cl
    self.value = v
    #print 'advanceFron', fr, upto, cl ,v
    return cl

  def getValue(self): return self.value


class IOrToCandidateBug(TestCase):
  def testToCandidateBug(self):
    s = IISet((0,))
    ss = IOr(IBTree(s), _ISearchCandidate1Hit2())
    ss.complete()
    self.assertEqual(ss.advanceFrom(None, None), HIT)
    self.assertEqual(ss.getValue(), 0)
    self.assertEqual(ss.advanceFrom(ss.getValue(), None), HIT)
    self.assertEqual(ss.getValue(), 2)
    self.assertEqual(ss.advanceFrom(ss.getValue(), None), END)



class TestCompletion(TestCase):
  set = [5, [1,2,3], 4, [None, 6], None]

  def testAnd(self):
    self._check('and')

  def testOr(self):
    self._check('or')

  def _check(self, op):
    search, flat = self._build(op, self.set)
    sizes = [s.getEstimatedSize() for s in search._searches if s is not None]
    proper = [i for i in flat if i is not None]
    no_improper = len(flat) - len(proper)
    if no_improper:
      self.assertEqual(sizes[-no_improper:], [None] * no_improper)
      sizes = sizes[:-no_improper]
    proper.sort()
    if op == 'or': proper.reverse()
    self.assertEqual(proper, sizes)

  def _build(self, op, set):
    class _Proxy(ISearch):
      def __init__(self, size): self._s = size
      def getEstimatedSize(self): return self._s
    sl = []; il = []
    for s in set:
      if isinstance(s, list):
        search, sizes = self._build(op, s)
        sl.append(search); il += sizes
      else: sl.append(_Proxy(s)), il.append(s)
    cl = op == 'and' and IAnd or IOr
    search = cl(*sl); search.complete()
    return search, il

    

def test_suite():
  return TestSuite([makeSuite(cl) for cl in [
                     IOrTests,
                     IOrCoverTests,
                     INotTests,
                     IFilterTests,
                     IntegrationTests,
                     IOrToCandidateBug,
                     TestCompletion,
                     ]
                    ]
                   )

def main():
  tester = TextTestRunner()
  tester.run(test_suite())

if __name__ == '__main__': main()

